﻿package {
	
	import flash.display.SimpleButton;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.display.MovieClip;
	import flash.display.Shape;
	import flash.geom.Point;
	import flash.ui.Keyboard;
	
	
	/**
	 * 1. フィールドに壁を描画する
	 * 2. フィールドにテトリミノを表示してみる
	 * 3. テトリミノを動かしてみる
	 * 
	 * @author necoEngine bamboo-satellites
	 * @version 0.0.2
	 * @see http://www.bamboo-satellites.com/
	 */
	public class main extends MovieClip {
		
		private const CELL_SIZE:uint      = 20; // セルのサイズ。正方形。
		private const FIELD_LEN_W:uint    = 12; // フィールドの完全幅。壁を除くと10。
		private const FIELD_LEN_H:uint    = 23; // フィールドの完全高さ。壁、テトリミノ出現領域分を除くと20。
		private const NUM_TETROMINOS:uint = 7;  // テトリミノの種類数
		
		private const EMPTY:uint      = 0;  // 空白コード
		private const TETROMINO1:uint = 1;  // テトリミノコード 棒形
		private const TETROMINO2:uint = 2;  // テトリミノコード　正方形
		private const TETROMINO3:uint = 3;  // テトリミノコード　S字
		private const TETROMINO4:uint = 4;  // テトリミノコード　Z字
		private const TETROMINO5:uint = 5;  // テトリミノコード　J字
		private const TETROMINO6:uint = 6;  // テトリミノコード　L字
		private const TETROMINO7:uint = 7;  // テトリミノコード　T字
		private const WALL:uint       = 99; // 壁コード
		
		private const LINE_COLOR:uint = 0x999999; // セルの線の色
		
		
		private var fieldData:Vector.<Vector.<int>>;      // フィールドの状態を常に保存しておく配列
		private var fieldContainer:Shape;                 // フィールドのデータを元に状態を描画する表示コンテナ。これ自体にイベントは必要ないのでShape。
		private var dropTetromino:Vector.<Vector.<uint>>; // 落下中のテトリミノのパターンデータ
		private var dropLocation:Point;                   // 落下中のテトリミノの位置情報
		
		
		public function main():void {
			
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
			
		} // constructor close
		
		
		
//================================================================================
// Init
//================================================================================
		private function init(e:Event = null):void {
			
			removeEventListener(Event.ADDED_TO_STAGE, init);
			// entry point
			
			// フィールドデータの初期化
			fieldData = fieldInit();
			
			
			//[テスト] フィールドデータにテトリミノを配置する
			var tetIndex:int = Math.floor(Math.random() * NUM_TETROMINOS) + 1;  // 新しいテトリミノの番号を選ぶ
			dropTetromino = tetriminoInit(tetIndex);
			dropLocation  = new Point(4, 3);
			fieldMapping(fieldData, dropTetromino, dropLocation); // フィールドデータにテトリミノのデータを反映する。仮配置座標x:4, y:3
			
			
			// フィールドの状態を描画する表示コンテナを生成して、配置。
			fieldContainer = new Shape();
			addChild(fieldContainer);
			fieldContainer.x = CELL_SIZE;
			fieldContainer.y = CELL_SIZE;
			// フィールド描画
			drawField(fieldData, fieldContainer);
			
			//ステージイベント
			stage.addEventListener(KeyboardEvent.KEY_DOWN, onKeyDownHandler);
			
			
			//[テスト] 更新ボタンを設置。
			var updateButton:SimpleButton = new ButtonCreater("update");
			addChild(updateButton);
			updateButton.x = stage.stageWidth  - updateButton.width  - CELL_SIZE;
			updateButton.y = stage.stageHeight - updateButton.height - CELL_SIZE;
			updateButton.addEventListener(MouseEvent.MOUSE_DOWN, onUpdate);
			
		} // init() close
		
		
		private function fieldInit():Vector.<Vector.<int>> {
			/**
			 * 機能：フィールドの初期化データ配列を返す
			 * 引数：なし
			 * 戻り値：Vector2次元配列
			 * 備考：最終的には初期描画処理も含める予定
			 */
			
			return Vector.<Vector.<int>>([
				Vector.<int>([EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY]),
				Vector.<int>([ WALL, WALL, WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL, WALL, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY,EMPTY, WALL]),
				Vector.<int>([ WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL, WALL])
			]);
			
		} // fieldInit() close
		
		private function tetriminoInit(code:uint):Vector.<Vector.<uint>> {
			/**
			 * 機能：引数で取ったテトリミノコードに対応したテトリミノのパターン配列を返す
			 * 引数：uint テトリミノのタイプを表すコード
			 * 戻り値：Vector2次元配列　テトリミノのパターンデータ。
			 * 備考：
			 */
			
			switch (code) {
				case TETROMINO1: // 棒形
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,     EMPTY,     EMPTY,     EMPTY]),
						Vector.<uint>([TETROMINO1,TETROMINO1,TETROMINO1,TETROMINO1]),
						Vector.<uint>([EMPTY,     EMPTY,     EMPTY,     EMPTY]),
						Vector.<uint>([EMPTY,     EMPTY,     EMPTY,     EMPTY])
					]);
				case TETROMINO2: // 正方形
					return Vector.<Vector.<uint>>([
						Vector.<uint>([TETROMINO2, TETROMINO2]),
						Vector.<uint>([TETROMINO2, TETROMINO2])
					]);
				case TETROMINO3: // S字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,      TETROMINO3, TETROMINO3]),
						Vector.<uint>([TETROMINO3, TETROMINO3, EMPTY]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				case TETROMINO4: // Z字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([TETROMINO4, TETROMINO4, EMPTY]),
						Vector.<uint>([EMPTY,      TETROMINO4, TETROMINO4]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				case TETROMINO5: // J字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([TETROMINO5, EMPTY,      EMPTY]),
						Vector.<uint>([TETROMINO5, TETROMINO5, TETROMINO5]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				case TETROMINO6: // L字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,      EMPTY,      TETROMINO6]),
						Vector.<uint>([TETROMINO6, TETROMINO6, TETROMINO6]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
				default: // T字
					return Vector.<Vector.<uint>>([
						Vector.<uint>([EMPTY,      TETROMINO7, EMPTY]),
						Vector.<uint>([TETROMINO7, TETROMINO7, TETROMINO7]),
						Vector.<uint>([EMPTY,      EMPTY,      EMPTY])
					]);
			}
			
		} // tetriminoInit() close
		
		private function colorInit(code:uint):uint {
			/**
			 * 機能：引数で取ったコードに対応したカラー値を返す
			 * 引数：uint
			 * 戻り値：uint 16進数
			 * 備考：
			 */
			
			switch (code) {
				case TETROMINO1: return 0x4DE6E1; // 棒形
				case TETROMINO2: return 0xE9EE11; // 正方形
				case TETROMINO3: return 0x46DF20; // S字
				case TETROMINO4: return 0xED2212; // Z字
				case TETROMINO5: return 0x2746D8; // J字
				case TETROMINO6: return 0xE49E1B; // L字
				case TETROMINO7: return 0xEA68E7; // T字
				default:         return 0x7A5B52; // 壁
			}
			
		} // colorInit() close
		
		
		
//================================================================================
// Tetromino Control
//================================================================================
		private function onKeyDownHandler(event:KeyboardEvent):void {
			/**
			 * 機能：キーボートイベントを検出して、キーに割り当てられた処理へ移行する。
			 * 引数：イベント
			 * 戻り値：なし
			 * 備考：拡張性の維持の為にキーボードイベントの検出、振り分けの処理のみに限定する。それ以外の動作処理は記入しない。
			 */
			
			switch(event.keyCode) {
				case Keyboard.LEFT:  tetriminoControl(Keyboard.LEFT,  fieldData, dropTetromino, dropLocation, fieldContainer); break;
				case Keyboard.RIGHT: tetriminoControl(Keyboard.RIGHT, fieldData, dropTetromino, dropLocation, fieldContainer); break;
				case Keyboard.DOWN:  tetriminoControl(Keyboard.DOWN,  fieldData, dropTetromino, dropLocation, fieldContainer); break;
			}
			
		} // onKeyDownHandler() close
		
		
		private function tetriminoControl(direct:uint, data:Vector.<Vector.<int>>, tetromino:Vector.<Vector.<uint>>, loc:Point, container:Shape):Boolean {
			/**
			 * 機能：指示に合わせてテトリミノを動かす。引数でとったコードに該当する処理を実行する。
			 * 引数：第1引数 テトリミノを動かす指示コード。基本的にはキーコード。
			 *      第2引数 フィールドデータ配列
			 *      第3引数 テトリミノパターンデータ
			 *      第4引数 テトリミノ座標
			 *      第5引数 フィールド表示コンテナ
			 * 戻り値： 真偽値
			 * 備考：
			 */
			
			switch (direct) {
				case Keyboard.LEFT: // 左
					if (getLeftHit(data, tetromino, loc)) return true; // 真が返る（左に何かある）と処理を終了
					fieldMapping(data, tetromino, loc, true); // 動かせる場合はフィールドデータから直前のテトリミノデータを削除
					loc.x--; // 位置情報の更新
					break;
					
				case Keyboard.RIGHT: // 右
					if (getRightHit(data, tetromino, loc)) return true; // 真が返る（右に何かある）と処理を終了
					fieldMapping(data, tetromino, loc, true); // 動かせる場合はフィールドデータから直前のテトリミノデータを削除
					loc.x++; // 位置情報の更新
					break;
					
				case Keyboard.DOWN: // 下
					if (getBottomHit(data, tetromino, loc)) return true; // 真が返る（下に何かある）と処理を終了
					fieldMapping(data, tetromino, loc, true); // 動かせる場合はフィールドデータから直前のテトリミノデータを削除
					loc.y++; // 位置情報の更新
					break;
			}
			
			fieldMapping(data, tetromino, loc); // フィールドデータにテトリミノの新しい位置情報を反映
			drawField(data, container);         // フィールドを再描画
			
			return false;
			
		} // tetriminoControl() close
		
		
		//------------------------------------------------------------
		// Tetromino Hit Check
		//
		// 機能：テトリミノパターン配列内のテトリミノコードが入った要素を探し、検出した場合はその要素のフィールドでの位置から隣接要素を参照して何かあれば真を返す。
		// 引数：
		// 戻り値：真偽値
		// 備考：とりあえず壁を検出すると終了
		//------------------------------------------------------------
		private function getLeftHit(data:Vector.<Vector.<int>>, tetromino:Vector.<Vector.<uint>>, loc:Point):Boolean {
			
			var x:int, y:int, fx:int, fy:int, len:uint = tetromino.length;
			for (x = 0; x < len; x++) { // 左から右へ走査
			for (y = 0; y < len; y++) { // 上から下へ走査
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy][fx - 1] == WALL) return true;
				}
			}
			}
			return false;
			
		} // getLeftHit() close
		
		private function getRightHit(data:Vector.<Vector.<int>>, tetromino:Vector.<Vector.<uint>>, loc:Point):Boolean {
			
			var x:int, y:int, fx:int, fy:int, len:uint = tetromino.length;
			for (x = len - 1; x >= 0; x--) { // 右から左へ走査
			for (y = 0; y < len; y++) {      // 上から下へ走査
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy][fx + 1] == WALL) return true;
				}
			}
			}
			return false;
			
		} // getRightHit() close
		
		private function getBottomHit(data:Vector.<Vector.<int>>, tetromino:Vector.<Vector.<uint>>, loc:Point):Boolean {
			
			var x:int, y:int, fx:int, fy:int, len:uint = tetromino.length;
			for (y = len - 1; y >= 0; y--) { // 下から上へ走査
			for (x = 0; x < len; x++) {      // 左から右へ走査
				if (tetromino[y][x]) {
					fx = loc.x + x; // テトリミノコードが入ったセルの、フィールド上でのx位置
					fy = loc.y + y; // テトリミノコードが入ったセルの、フィールド上でのy位置
					if (data[fy + 1][fx] == WALL) return true;
				}
			}
			}
			return false;
			
		} // getBottomHit() close
		
		
		
//================================================================================
// Field
//================================================================================
		private function fieldMapping(data:Vector.<Vector.<int>>, tetromino:Vector.<Vector.<uint>>, loc:Point, clear:Boolean = false):void {
			/**
			 * 機能：フィールドデータ配列に、出現したテトリミノのパターンデータを反映する。
			 * 引数：第1引数 Vector二次元配列 フィールドデータ配列
			 *      第2引数 Vector二次元配列 出現したテトリミノのパターンデータ配列
			 *      第3引数 uint 出現したテトリミノの位置
			 *      第4引数 テトリミノコードの削除指示
			 * 戻り値：なし
			 * 備考：現在はとりあえずの最小処理。
			 */
			
			var x:uint, y:uint, ex:uint, ey:uint, px:uint, py:uint, code:uint, tetSize:uint = tetromino.length;
			
			ex = loc.x + tetSize; // x軸走査の終了位置
			ey = loc.y + tetSize; // y軸走査の終了位置
			for (y = loc.y; y < ey; y++) {
			for (x = loc.x; x < ex; x++) {
				px = x - loc.x; // テトリミノパターンデータ配列用のインデックス
				py = y - loc.y; // テトリミノパターンデータ配列用のインデックス
				code = tetromino[py][px];
				if (code) {
					data[y][x] = (clear)? EMPTY: code;
				}
			}
			}
			
		} // fieldMapping() close
		
		
		private function drawField(data:Vector.<Vector.<int>>, container:Shape):void {
			/**
			 * 機能：フィールドデータの情報に基ずいて、表示コンテナにフィールドを描画する。
			 * 引数：第1引数・フィールドデータ
			 *      第2引数・描画する表示コンテナ
			 * 戻り値：なし
			 * 備考：
			 */
			
			var x:int, y:int, px:int, py:int, code:int;
			
			container.graphics.clear();
			
			for (y = 0; y < FIELD_LEN_H; y++) {
				px = 0;
				for (x = 0; x < FIELD_LEN_W; x++) {
					code = data[y][x];
					if (code > EMPTY) {
						container.graphics.beginFill(colorInit(code));
						container.graphics.lineStyle(1, LINE_COLOR);
						container.graphics.drawRect(px, py, CELL_SIZE, CELL_SIZE);
						container.graphics.endFill();
					}
					px += CELL_SIZE;
				}
				py += CELL_SIZE;
			}
			
			container.graphics.endFill();
			
		} // drawField() close
		
		
		
//================================================================================
// Develop
// 
// 開発中に使用するテスト用メソッド。最終的には削除。
//================================================================================
		private function onUpdate(event:MouseEvent):void {
			
			updateTetromino();
			
		} // onUpdateTetromino() close
		
		
		private function updateTetromino():void {
			/**
			 * 機能：フィールドに配置されたテトリミノの削除（フィールドデータ上のテトリミノデータを空白にする）し、新しいテトリミノのデータをマッピング、再描画までを実行する。
			 * 引数：なし
			 * 戻り値：なし
			 * 備考：テスト用なので割りと適当
			 */
			
			//fieldData配列の要素を明示的に破棄しておく
			fieldData.forEach(function (element:Vector.<int>, index:int, array:Vector.<Vector.<int>>):void { element = null; } );
			fieldData = null;
			//テトリミノパターンデータ配列の要素を明示的に破棄しておく
			dropTetromino.forEach(function (element:Vector.<uint>, index:int, array:Vector.<Vector.<uint>>):void { element = null;} );
			dropTetromino = null;
			// 位置情報破棄
			dropLocation = null;
			
			// フィールドデータ初期化
			fieldData = fieldInit();
			
			// 新しいテトリミノを生成
			var tetIndex:int = Math.floor(Math.random() * 7) + 1;
			dropTetromino = tetriminoInit(tetIndex);
			dropLocation  = new Point(4, 3);
			
			// 新しいテトリミノをフィールドデータに反映する
			fieldMapping(fieldData, dropTetromino, dropLocation); // 仮配置座標x:4, y:3
			
			// 更新したフィールドデータを下にフィールドを再描画
			drawField(fieldData, fieldContainer);
			
		} // () close
	}
}